/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (collect.C)
  
  Copyright:  Damian J. J. Farnell, Mar2000

  Updated:    Joerg Schulenburg 2004-2009

  ===================================================

  ---------------------------------------------------

  PROGRAM NAME: collect.C 

  PURPOSE: collect same configurations to reduce loops

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdlib.h>
#include <stdio.h>
// #include <math.h>
 #include <string.h>

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"
// #include "nr.h"
// #include "nrutil.h"
#include "storage.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

#ifndef DO_COLLECT
// if set terms with same indices will be summed (sq-8: size_oterms factor 1/6, time+5%)
#define DO_COLLECT 1  /* set to 0 to switch off collect, default = 1 */
#endif

/* Joerg05:
 * who is calling  collect_term?
 *  62 times by obtain()+term2()..term62() in obtain.C 
 * parameters and its ranges?
 *  indices[0] - configuration (lattice animal), 1..NumberConfigurations
 *                                          new: 0 for GS Energy
 *  indices[1] - term 2..62, STOP_TERM=0=end
 *  indices[2..5] -  0,1..NumConfig ??  (0,0,0,0)=not_used(check it)
 *  dummy = CCM coefficient (old: CCMcoeff)
 * ToDo: exchange [0] <-> [1]
 * Outputfile fp1=*.terms is used for solve()
 * Outputfile fp1=*_gs    is used for ...     (new Sep05) ???
 * maximum number of terms and clusters for sample models?
 *  MAX_NUMBER_TERMS_IN_CCM_EQNS to big? (*46Bytes)
 * ToDo: dynamic arrays
 */

// MAX_NUMBER_TERMS_IN_CCM_EQNS renamed to MAXNUM_HASHED_TERMS
// MAXNUM_HASHED_TERMS should be a prime number for good hashing
// static arrays: MAXNUM_HASHED_TERMS * 44 byte, dfl=(1<<20)*36B=36MB (old)
// static arrays: MAXNUM_HASHED_TERMS * 36 byte, dfl=(1<<22)*36B=144MB (new)
// #define MAXNUM_HASHED_TERMS ((1<<20)-3)  // JS1002 obsolete, new=hash_max
// may be set by option (input.c) or nC-table-estimation (define.c)
//   smaller values will speedup computation (CPU/memory cashing)
int hash_bits=20;     // maximum is 31 (= int.bits)
                      // min=26 bits for s=1/2 LC-18-18

// we could save hash space when we order terms by num factors [0..4]
//  v1002 does buffer all factor lengths in same buffer 
// s05-lc-6-6   nC=3+5+1=9         T=119   hb= 6 buff=35   bcfg=9  2[4]+14*0+8[3]+11[2]
//  partly sum to zero: 09.32 = -09.31
// s05-lc-6-8   nC=4+13+8=25       T=937   hb= 7 buff=102  bcfg=12 4[4]+12*0+33[3]+2*0+51[2]
// s05-lc-8-8   nC=4+13+8+1=26     T=1120  hb= 9 buff=295  bcfg=26 53[4]+40*0+157[3]+45[2]
// s05-lc-8-10  nC=5+27+35+13=80   T=11441 hb=10 buff=802  bcfg=75 ...
// s05-lc-8-16  nC=8+118+616+1269=2011 T=1723981 hb=12 buff=1565 bcfg=610 ...
// s05-lc-10-10 nC=5+27+35+13+1=81 T=13063 hb=12 buff=3048 bcfg=81 ...
// s05-lc-10-20 nC=10+235+2220+=27422
// s05-lc-18-18 MAXHASH=1<<25(size=1.2GB)=collis. (MAXHASH=1<<26=ld-error 2GB)
// # unbalanced because of small number of connected long configs !!! 
//   T propto nMaxlenC
// max. possible buffer reduction is only 3? buff is of order T?
//  term09.34[4] 6n+2=2+2+2+2 buffmax=3*2*1=6 or 4*3*2*1=24
//               8n+2=2+2+2+4 buffmax=13*4*3*2=312
//
// ToDo: measure CPU+memory balance max_stored vs. sum_stored/threads
// ToDo: move this table to obtain.C and ...
// ToDo: add termGSE or use term13 + debugstring like below
//  ...  use as debug arguments (...,1(="zz"),"function")
//   this table is wrong for id=3 but valid for id=1,2
//  xxzLC, xxzSQ: SzSz=1.2..1.5 + SxSy+SySx=8.26+9.27..9.45
// JS0910 ToDo: decouple F,H,M product from Sz,S+ product (like Sven)
const char *term62s[62+2]={"",         // 0.0  ToDo: reduce 62 by 10 k_to_m's?
//  Hkm is sum over Si+ ... see diss sven
//     term13[3]      term62[20] numcoeffs[5] 
       "zz                  ??? ?[0]", // 1.1 not used ???
       "zz           Fm.S+m.Szk  [1]", // 1.2        xxzLC(z) isingLC
       "zz           Fk.S+k.Szm  [1]", // 1.3 = ~1.2 ... merged
       "zz          Hkm.S+k.S+m  [1]", // 1.4        ...
       "zz        Fk.Fm.S+k.S+m  [2]", // 1.5        xxzLC(z) isingLC
       "z+              S+m.Szk  [0]", // 2.6        kago_triangle? lc_mf? MB
       "z+           Fk.S+k.S+m  [1]", // 2.7
       "z-        -2*Fm.Szk.Szm  [1]", // 3.8 cancels 5.17? kago_triangle?
       "z-       -2*Hkm.S+k.Szm  [1]", // 3.9
       "z-       -Fm.Fm.S+m.Szk  [2]", // 3.10
       "z-    -2*Hkm.Fm.S+k.S+m  [2]", // 3.11
       "z-     -2*Fk.Fm.S+k.Szm  [2]", // 3.12
       "z-    -Fk.Fm.Fm.S+k.S+m  [3]", // 3.13 
// ToDo: collect +z and z+ in case of directional=0 ???
       "+z              S+k.Szm  [0]", // 4.14 = ~2.6 merged
       "+z           Fm.S+k.S+m  [1]", // 4.15 = ~2.7 merged
       "-z       -2*Hmk.S+m.Szk  [1]", // 5.16 = ~3.9 merged kago_triangle?
       "-z        -2*Fm*Szk*Szm  [1]", // 5.17 = ~3.8 merged
       "-z     -2*Fm.Fk.S+m.Szk  [2]", // 5.18 = ~3.12 merged
       "-z    -2*Hkm.Fk.S+k.S+m  [2]", // 5.19 = ~3.11 merged
       "-z       -Fk.Fk.S+k.Szm  [2]", // 5.20 = ~3.10 merged
       "-z    -Fk.Fk.Fm.S+k.S+m  [3]", // 5.21 = ~3.13 merged
       "+-        -2*Fm.S+k.Szm  [1]", // 6.22 CaVO_neel? j1j2_aus_srcubo kago
       "+-       -Fm.Fm.S+k.S+m  [2]", // 6.23
       "-+        -2*Fk.S+m.Szk  [1]", // 7.24 = ~6.22 merged
       "-+       -Fk.Fk.S+k.S+m  [2]", // 7.25 = ~6.23 merged
       "++              S+k.S+m  [0]", // 8.26  xxzLC(xy)
       "--        4*Hkm.Szk.Szm  [1]", // 9.27  xxzLC(xy)
       "--    2*Hkm.Hkm.S+k.S+m  [2]", // 9.28  ...
       "--     4*Hkm.Fm.S+m.Szk  [2]", // 9.29
       "--      4*Fk.Fm.Szk.Szm  [2]", // 9.30
       "--   2*Fm.Fm.Fk.S+m.Szk  [3]", // 9.31 S=3/2-LC
       "--  4*Fk.Fm.Hkm.S+k.S+m  [3]", // 9.32 S=3/2-LC    slow!
       "--   2*Fk.Fk.Fm.S+k.Szm  [3]", // 9.33 = ~9.31 merged
       "--  Fk.Fk.Fm.Fm.S+k.S+m  [4]", // 9.34 S=3/2-LC (biggest contribution CPUh)
       "--     4*Hkm.Fk.S+k.Szm  [2]", // 9.35 = ~9.29 merged
       "--       2*Mkkm.S+k.Szm  [1] 2s>1", // 9.36
       "--       2*Mkmm.S+m.Szk  [1] 2s>1", // 9.37 = ~9.36 merged
       "--        Nkkmm.S+k.S+m  [1]", // 9.38
       "--      Fm.Mkkm.S+k.S+m  [2]", // 9.39
       "--     2*Fk.Hmm.S+m.Szk  [2]", // 9.40                xxLC-3/2
       "--      Fk.Mmmk.S+k.S+m  [2]", // 9.41 = ~9.39 merged ...
       "--     2*Fm.Hkk.S+k.Szm  [2]", // 9.42 = ~9.40 merged ...
       "--      Hkk.Hmm.S+k.S+m  [2]", // 9.43                ...
       "--    Hkk.Fm.Fm.S+k.S+m  [3] 2s>1", //  9.44          xxLC-3/2 ... slow
       "--    Hmm.Fk.Fk.S+k.S+m  [3] 2s>1", //  9.45 = ~9.44 merged xxzLC(xy)
       "z-         -Hmm.S+m.Szk  [1] 2s>1", //  3.46
       "z-        -Mkmm.S+k.S+m  [1] 2s>1", //  3.47
       "z-      -Fk.Hmm.S+k.S+m  [2] 2s>1", //  3.48
       "-z        -Mkkm.S+k.S+m  [1] 2s>1", //  5.49 = ~3.47 merged
       "-z         -Hkk.S+k.Szm  [1] 2s>1", //  5.50 = ~3.46 merged
       "-z      -Fm.Hkk.S+k.S+m  [2] 2s>1", //  5.51 = ~3.48 merged
       "+-         -Hmm.S+k.S+m  [1] 2s>1", //  6.52  ToDo: sample system?
       "-+         -Hkk.S+k.S+m  [1] 2s>1", //  7.53 = ~6.52 merged
       "z                Fk.S+k  [1]",      // 10.54  lc_mf
       "-             -2*Fk.Szk  [1]",      // 11.55  lc_mf
       "-              -Hkk.S+k  [1]",
       "-            -Fk.Fk.S+k  [2]",
       "z^2         2Fk.S+k.Szk  [1]",      // 12.58
       "z^2          Fk.S+k.S+k  [1]",      // 12.59
       "z^2         Hkk.S+k.S+k  [1]",
       "z^2       Fk.Fk.S+k.S+k  [2]",
       "+                   S+k  [0]",      // 13.62  lc_mf
       ""};
// difference? P = F for excited states? Q = H for excited states?
//   Sven uses same subroutines (JS0910)
const char *term81s[81+2]={"",         // 0.0
//     term13[3] term81[20] numcoeffs[5] 
       "zz           Pi.S+i.Szj  [1]", // 1.1
       "zz        Pi.Fj.S+i.S+j  [2]", // 1.2
       "zz           Pj.S+j.Szi  [1]", // 1.3 = ~1.1
       "zz        Pj.Fi.S+i.S+j  [2]", // 1.4 = ~1.2
       "zz          Qij.S+i.S+j  [1]", // 1.5
       "z+           Pi.S+i.S+m  [1]", // 2.6
       "z-     -2*Pi.Fj.S+i.Szj  [2]", // 3.7
       "z-      -Pi.Gjj.S+i.S+j  [2]", // 3.8
       "z-    -Pi.Fj.Fj.S+i.S+j  [3]", // 3.9
       "z-         -2Pj.Szi.Szj  [1]", // 3.10
       "z-     -2*Pj.Fj.S+j.Szi  [2]", // 3.11
       "z-     -2*Pj.Fi.S+i.Szj  [2]", // 3.12
       "z-    -2*Pj.Gij.S+i.S+j  [2]", // 3.13
       "z-  -2*Pj.Fi.Fj.S+i.S+j  [3]", // 3.14
       "z-       -2*Qij.S+i.Szj  [1]", // 3.15
       "z-    -2*Qij.Fj.S+i.S+j  [2]", // 3.16
       "z-         -Qjj.S+j.Szi  [1]", // 3.17
       "z-      -Qjj.Fi.S+i.S+j  [2]", // 3.18
       "z-        -Lijj.S+i.S+j  [1]", // 3.19
       "+z           Pj.S+i.S+j  [1]", // 4.20 like 2.6 backwards
       "-z     -2*Pj.Fi.S+j.Szi  [2]", // 5.21 like 3.7 backwards
       "-z      -Pj.Gii.S+i.S+j  [2]", // 5.22 like 3.8 backwards
       "-z    -Pj.Fi.Fi.S+i.S+j  [3]", // 5.23 like 3.9 backwards
       "-z        -2*Pi.Szi.Szj  [1]", // 5.24 like 3.10 backwards
       "-z     -2*Pi.Fj.S+j.Szi  [2]", // 5.25 like 3.12 backwards
       "-z     -2*Pi.Fi.S+i.Szj  [2]", // 5.26 like 3.11 backwards
       "-z    -2*Pi.Gij.S+i.S+j  [2]", // 5.27 like 3.13 backwards
       "-z  -2*Pi.Fi.Fj.S+i.S+j  [3]", // 5.28 like 3.14 backwards
       "-z       -2*Qij.S+j.Szi  [1]", // 5.29 like 3.15 backwards
       "-z    -2*Qij.Fi.S+i.S+j  [2]", // 5.30 like 3.16
       "-z         -Qii.S+i.Szj  [1]", // 5.31 like 3.17
       "-z    -2*Qii.Fj.S+i.S+j  [2]", // 5.32 like 3.18
       "-z        -Liij.S+i.S+j  [1]", // 5.33 like 3.19
       "+-        -2*Pj.S+i.Szj  [1]", // 6.34
       "+-     -2*Pj.Fj.S+i.S+j  [2]", // 6.35
       "+-         -Qjj.S+i.S+j  [1]", // 6.36
       "-+        -2*Pi.S+j.Szi  [1]", // 7.37 like 6.34
       "-+     -2*Pi.Fi.S+i.S+j  [2]", // 7.38 like 6.35
       "-+         -Qii.S+i.S+j  [1]", // 7.39 like 6.36
       "--      4*Pi.Fj.Szi.Szj  [2]", // 9.40
       "--     4*Pi.Gij.S+i.Szj  [2]", // 9.41 like 9.62 but forward
       "--     2*Pi.Gjj.S+j.Szi  [2]", // 9.42
       "--    2*Pi.Hijj.S+i.S+j  [2]",
       "--   4*Pk.Fk.Fm.S+k.Szm  [3]",
       "--  4*Pk.Fm.Hkm.S+k.S+m  [3]", // 9.45   = ~9.70???
       "--  2*Pi.Fi.Gjj.S+i.S+j  [3]", // 9.46 slow
       "-- 2Pk.Fk.Fm.Fm.S+k.S+m  [4]", // 9.47 = ~9.69,  slowest !!!
       "--   2*Pk.Fm.Fm.S+m.Szj  [3]", // 9.48
       "--     2*Qkk.Fj.S+i.Szj  [2]",
       "--      Qkk.Hmm.S+k.S+m  [2]",
       "--    Qkk.Fm.Fm.S+k.S+m  [3]", // 9.51
       "--        4*Qkm.Szk.Szm  [1]", // 9.52
       "--     4*Qkm.Fm.S+m.Szk  [2]", // 9.53
       "--     4*Qij.Fi.S+i.Szj  [2]", // 9.54
       "--    4*Qkm.Hkm.S+k.S+m  [2]", // 9.55
       "--  4*Qkm.Fk.Fm.S+k.S+m  [3]", // 9.56
       "--       2*Lkkm.S+k.Szm  [1]",
       "--    2*Lkkm.Fm.S+k.S+m  [2]",
       "--       2*Lijj.S+j.Szi  [1]", // 9.59 NOTE: DIRECTIONALITY TO BE ADDED FOR TRIANGLE!
       "--    2*Lmmk.Fk.S+k.S+m  [2]",
       "--        Mkkmm.S+k.S+m  [1]",
       "--     4*Pj.Gij.S+j.Szi  [2]", // 9.62 like 9.41 but backwards
       "--    2*Pj.Liij.S+i.S+j  [2]",
       "--      4*Pj.Fi.Szi.Szj  [2]", // most time consuming?
       "--   4*Pj.Fi.Fj.S+j.Szi  [3]", // 9.65 same as 9.44 but km <=> ji ???
       "--     2*Pj.Gii.S+i.Szj  [2]", // 9.66 = ~9.42 xxzLC S=3/2
       "--  2*Pj.Fj.Gii.S+i.S+j  [3]", // 9.67 = ~9.46
       "--   2*Pm.Fk.Fk.S+k.Szm  [3]", // 9.68 = ~9.48
       "-- 2Pm.Fm.Fk.Fk.S+k.S+m  [4]", // 9.69 = ~9.47 xxzLC S=3/2
       "--  4*Pj.Fi.Gkm.S+k.S+m  [3]", // 9.70 = ~9.45??? slow
       "--     2*Qjj.Fi.S+j.Szi  [2]",
       "--      Qjj.Hii.S+i.S+j  [2]",
       "--    Qjj.Fi.Fi.S+k.S+m  [3]", // 9.73 = ~9.51
       "z                Pk.S+k  [1]", // 10.74
       "-             -2*Pk.Szk  [1]", // 11.75
       "-          -2*Pk.Fk.S+k  [2]",
       "-              -Qkk.S+k  [1]",
       "z^2        2*Pk.S+k.Szk  [1]", // 12.78
       "z^2          Pk.S+k.S+k  [1]",
       "z^2         Qkk.S+k.S+k  [1]",
       "z^2     2*Pk.Fk.S+k.S+k  [2]",
       ""};
const char *term13s[13+2]={"",  //  9pairs+4single = 6(+3back)pairs+4single  
       "zz ", // [0..2] 1 ??? =gse1  (1..9 are bonds)
       "z+ ", // [0..1] 2
       "z- ", // [1..3] 3     =gse4
       "+z ", // [0..1] 4 = ~2
       "-z ", // [1..3] 5 = ~3
       "+- ", // [1..2] 6
       "-+ ", // [1..2] 7 = ~6
       "++ ", // [0   ] 8
       "-- ", // [1..4] 9  = gse2
       "z  ", // [1   ] 10     = gse6???  (10..13 are single site terms)
       "-  ", // [1..2] 11 = gse7???
       "z^2", // [1..2] 12 = gse8=zz???
       "+  ", // [0   ] 13
       ""};

#define MAX_out 40            // reduced 1000 to 200 Mar08

/* export for measure speed */
unsigned long
     nterms_out0=0, /* count output terms summed to zero (not stored) */
     nterms_out1=0, /* count output terms per thread (summed at end and stop output) */
     nterms_out2=0, /* overflow */
     nterms_in1=0,  /* count input  terms per thread (summed at end) */
     nterms_in2=0;  /* overflow */
int flag_out[6]={1,1,1,1,1,1}; // set 0 after first run, reduce output!

#if DO_COLLECT
// gpg --gen-prime 1 31
#define HASH_PRIME 0x701F35F3   // not used @ moment

// work:
//  indices will be sorted (not used in the calling routine), Joerg Sep05
//  same or permutated indices[2..5] will be stored as sum only once
//  indices pointing to configurations 1..NumberConfigurations
//  permutated indices dont have the same probability! that means: (ToDo)
//       its difficult to simplify (no easy symmetry)
//       but it looks like that every term appears twice (ToDo: simplify!)
//
// id: 0=gs_terms, 1=terms 2=expectations 3=excitation
//
void collect_term(int indices[6], double dummy, int id) 
{
  int  k; long i, j; // , nio=7;
  static double na_in=0, na_out=0; /* Apr08 long to dbl, no bond sum */
  static int  id_old=-1;         /* flag for new collection */
  static long counter=0;         /* buffer index for desciptors,multipliers */
  static long max_buff=0,     /* maximum hash buffer size */
              max_bcfg=0;     /* config generating max_buff */
  /* hashed descriptors and multipliers {src_idx, term, dest_idx1..4}
   * in princip we dont need to store src_idx + term for every tuple
   * that would save 8 of 32 bytes but generate some extra code (done JS091013)
   * indices[0] = cfg + indices[1] = term not need to be hashed!
   */
  static long int hash_max=0; // JS1002: old MAXNUM_HASHED_TERMS=(1<<20)-3
  static int    hash_cfg  = -1;
  static int    hash_term = -1;
  static int    (* hash_descr)[4]=NULL; // 16 byte (v0909: 6*4B)
  static double  *hash_mult      =NULL; //  8 byte
  // it is very slow to search a big table for an entry sequentialy
  // so we introduce a hash table of indizes to that entries
  //  and a table of indices to that table to have fast sequential access 
  //  to spare filled hash table (Joerg Sep05, xxzLC-14 5x faster)
  // number of hashed entries is in counter
  // JS ToDo: instead of hash we could store a (binary) tree using
  //   sorted dest_descr1-4, lenCfg1-4
  // better of course is to collect in the generator functions (obtain)
  static long *hash=NULL; // 8B hashed index to entries
  static int  *hidx=NULL; // 4B index to hash entries
  static int hcol=0;   // counter for hash collisions, just as info
  static int hhit=0;   // counter for hash hits
  static int hnew=0;   // counter for new hash entries
  static int hval=-1;  // hash value
  static int nflush=0; // number of hash table overflows + flushs
  // count number of terms, only verbose output
  static unsigned long
    numiterms[81+1]={0}, // input terms from obtain, 81 or 62 for id=0..2, ...
    numoterms[14*5]={0}; // counting for each term13=1..13 for 0..4-coefficients
                         // ToDo: add index for max_cfg_len or cfg0_len?
  long maxterms_out=0, maxterms_in=0;
  const int *map;
  static int tmp1[6]={0,0,0,0,0,0}; // buffer for term_store()

  if (!Control.file) return;

  if (!hash_max) { // initialize memory once per run, JS1002
    hash_max=(1l<<hash_bits)-3;
    // JS1002: we alloc dynamic now to set hash_max at runtime (free on exit)
    //  4*4+8+4or8+4 = 32 or 36 Byte per hash entry
    hash_descr = (int (*)[4]) malloc((hash_max+1)*sizeof(int)*4 );
    hash_mult  = (double *)   malloc((hash_max+1)*sizeof(double));
    hash       = (long *)     malloc((hash_max+1)*sizeof(long)  );
    hidx       = (int *)      malloc((hash_max+1)*sizeof(int)   );
    if ((!hash_descr) || (!hash_mult) || (!hash) || (!hidx)) ERR("hash_mem");
    if (!myid) printf("collect.hash.ini: bits= %3d  max= %12ld\n\n",
                     hash_bits, hash_max);
  }

  map=((id==3)?map81:map62); // choose the right mapping

  // initialize for every bond or caller=gse/erg/exc/exp
  if (nterms_in1==0 && nterms_in2==0 && id_old!=id) {  // new file
    for (i=0;i<=81; i++) numiterms[i]=0; // initialize counter
    for (i=0;i<14*5;i++) numoterms[i]=0; // initialize counter
  }
  // GSE terms are not distributed over tasks, count once
  if (id!=0 || myid==0) { nterms_in1++; if(!nterms_in1) nterms_in2++; }

  if (hval<0) { hval=0; // used as flag, initialize hash once!
    for (i=0;i<hash_max+1;i++) { hidx[i]=hash[i]=0; }
  }

  if (counter){  // something stored into buffer
      /* save collected terms and reset buffer for next term */

    // check if last num_cfg differs fron this num_cfg 
    // term09.31,32,34 can not inter-collected, but 9.31+9.33 can collect
    //   should be done at obtain (do generate once only)
    //   hash table grows fast = o_term[4]/nCmaxlen (LC16=30e6=2^25,LC18=770e6=2^30)
    if (    hash_cfg < indices[0]  /* next configuration index */
       // indices[1]: term
       || indices[1] == STOP_TERM /* JS-2010-04-15 flush buffer condition */
       || map[hash_term] != map[indices[1]]  /* next term? 2..62, Apr08 JS */
// ToDo JS1002: we assume that same term13s on same type of bond have the
//  same coefficient! but cavo_striped.ccm has not! try without collect!
//     ||     hash_term  !=     indices[1]   /* next term? 2..62, Apr08 JS */
//     ||    0  > indices[0] /* flush buffer condition, obsolete JS-2010-04 */
      /* Sept 05: old was 0>i[0] but collect_gse needs 0>=i[0] */
//       ||    NumberConfigurations < indices[0]  # not id=3
       || counter >= hash_max) /* buffer full */
    {
        if (       hash_max == counter) {
           if (nflush<10 && flag_out[id]) {
             printf("HASH-Overflow happens for task= %4d cfg= %6d term%02d.%02d\n",
               myid, indices[0], map[indices[1]], indices[1]);
             fflush(stdout);
           }
           if (++nflush==10 && flag_out[id] && !myid)
            printf("excessive output is stopped after %d prints\n",MAX_out);
        }
	// first save information 
	for(j=1;j<=counter;j++){
	  
          // if (!hash_mult[j]) { nterms_out0++; continue; }  // summed to zero
          if (hash_mult[j]) {
            tmp1[0]=hash_cfg;
            tmp1[1]=hash_term;
            tmp1[2]=hash_descr[j][0];
            tmp1[3]=hash_descr[j][1];
            tmp1[4]=hash_descr[j][2];
            tmp1[5]=hash_descr[j][3];
            term_store(id, tmp1, hash_mult[j]);
            nterms_out1++; if(!nterms_out1) nterms_out2++;
          } else nterms_out0++;  // summed to zero
          if (Control.output<3 && !hash_mult[j]) continue; // no output
          
          // for (i=0;hash_descr[j][i] && i<4;i++); // ups on bounds checking
          for (i=0; i<4 && hash_descr[j][i]; i++); // SK1003+bgcc
          numoterms[ (map[hash_term]) * 5 + i ]++;
	  // printf out information
          if (flag_out[id])
	  if (Control.output > 1 && myid == 0
	    && nterms_out1 < MAX_out * (1UL<<Control.output)) {
	    printf("OUT: Cfg=%3d Term%02d    CCM= %4d %4d %4d %4d"
		   " % 7.3f %s [%d] j= %6d\n",
		   hash_cfg, map[hash_term],
		   hash_descr[j][0], hash_descr[j][1],
                   hash_descr[j][2], hash_descr[j][3], hash_mult[j],
                          term13s[map[hash_term]], (int)i, (int)j);
            if (nterms_out1+1==MAX_out * (1UL<<Control.output))
              printf("OUT: no output after MAX_out=%d terms\n", MAX_out);
          }
          hash[ hidx[j] ]=0; // reset hash entry
          hidx[j]=0; // not necessary
	  // now overwrite the old stored information 
	}
        for (i=1;i<=counter;i++) { hash[ hidx[i] ]=0; }
	counter=0;
      }
  }

  // ToDo: print OUT term description
  // i[2]>0 and i[3]=0 1st order coeff.
  // i[3]>0 and i[4]=0 2nd order coeff.
  // i[4]>0 and i[5]=0 3th order coeff.
  // i[5]>0            4th order coeff.
  if (flag_out[id])
  if (Control.output > 2 && nterms_out1 < MAX_out)
    printf("IN:  Cfg=%3d Term%02d.%02d CCM= %4d %4d %4d %4d"
           " % 7.3f %.23s\n",
       indices[0], map[indices[1]],
       indices[1],
       indices[2], indices[3],
       indices[4], indices[5], dummy,
       ((id==3)?term81s[indices[1]]:term62s[indices[1]]));
  if (flag_out[id])
  if (Control.output > 2 && myid == 0 && nterms_out1+1 == MAX_out)
    printf("IN:  Output stopped\n");

  // the last config has lot of terms, try to reduce further by
  // reducing all possible permutations to one config by sorting (big first)

  // for id=3 (excitations):
  //   x[2]    <= excitation coefficients
  //   x[3..5] <= g-s ket-state coefficients (NO PARTICULAR ORDER)
  // if (id!=3)  // Joerg Jun08: sorting for Excitad states failes
  for(;;) { // bubble sort indices
    j=1;
    for(i=((id==3)?3:2); i<5; i++) // for term_exc sort 3..5 only
      if (indices[i]<indices[i+1]) {  // exchange two indices
        j=0; k=indices[i+1]; indices[i+1]=indices[i]; indices[i]=k; 
      }
    if (j) break;
  }

  // Oct05: new hash table search (fast)
  j=0;
  if (indices[0]>=0 /* old */ && indices[1]!=STOP_TERM /* 2010-04 */)
  {
    /* ToDo: dont need to hash indices[0]! JS-2010-04 */
    hval=(((((                          indices[5])
              *(NumberConfigurations+1)+indices[4])
              *(NumberConfigurations+1)+indices[3])
              *(NumberConfigurations+1)+indices[2])
              *(NumberConfigurations+1)+indices[0]);
    if (id==3)
    hval=(((((                                 indices[5])
              *(NumberExcitedConfigurations+1)+indices[4])
              *(NumberExcitedConfigurations+1)+indices[3])
              *(NumberExcitedConfigurations+1)+indices[2])
              *(NumberExcitedConfigurations+1)+indices[0]);
    // try up to 1024 times,
    for(i=0;;i++) {
      hval = hval + indices[2] + 1;  // avoid overflow to negativ values
      // abs(hval) will fail for INT32_MIN
      if (hval<0) hval=~hval; // INT32_MIN..-1 <-> INT32_MAX..0
      hval %= hash_max; // MAX should be prime
      j = hash[hval];
      if (j>counter || j<0 || hval<0) { 
         printf("Bug86757 j=%ld counter=%ld hash=%d\n",j,counter,hval);
         exit(1);
      }
      if (j==0) {      // empty entry
             hash[hval]=counter+1;  // set hash entry
             hidx[counter+1]=hval;  // set sequential index to entry
             j=counter+1;           // store to table
             hnew++; break;
      } else {
        if ( indices[0]==hash_cfg &&
         map[indices[1]]==map[hash_term] &&  // term
             indices[2]==hash_descr[j][0] &&
             indices[3]==hash_descr[j][1] &&
             indices[4]==hash_descr[j][2] &&
             indices[5]==hash_descr[j][3]
           ) {
            hash_mult[j]+=dummy;  // hashed == searched
            hhit++; break;
         } else if (i>=1024) {
           static int hcol_out=1;
           hcol++; nterms_out1++;    // collision, write through
           if (!nterms_out1) nterms_out2++;
           { int i=0;
             for (i=0;hash_descr[j][i] && i<4;i++);
             numoterms[ (map[indices[1]]) * 5 + i ]++;
           }
           term_store(id,indices,dummy);
           if (hcol_out==hcol && flag_out[id]) { // avoid excessiv output
             if (hcol_out<<1) hcol_out<<=1;  // *=2
             if (!myid) // avoid excessiv output JS1002
             printf("%4d: collision no %8d cfg= %5d term= %3d %7.3f%%\n",
                     myid, hcol, indices[0], indices[1],
                     hcol*100./hash_max);
           }
           // ToDo: create/access a 2nd level hash table using 4*hashsize
           //       or use a tree (num descriptorindizes, terms), count!
           j=0; break;
         }
      }
    }
  }
  
//  if (0<=indices[0]) { // valid (ToDo: always? 2010-04)
  if (STOP_TERM!=indices[1]) { // valid (JS-2010-04)
    if (indices[1]<0 || indices[1]>((id==3)?81:62)) { 
      printf("Error collect.C L%d: term=%d\n",__LINE__,indices[1]); ERR(1);
    }
    numiterms[indices[1]]++; // for statistics only
  }

  if (/* old: 0<=indices[0], new 2010-04: */ indices[1]!=STOP_TERM)
  if ( j>counter /* && indices[1] != 0 */ ) {
    counter++;  /* potential overflow! */
    if (counter>max_buff){ /* get a better feeling about most terms (speed) */
      max_buff=counter; /* max hash size info */
      max_bcfg=indices[0]; }                 
    hash_cfg =indices[0]; // src cluster index
    hash_term=indices[1]; // term 1..62..82
    hash_descr[counter][0]=indices[2]; // dest cluster index 1..4
    hash_descr[counter][1]=indices[3];
    hash_descr[counter][2]=indices[4];
    hash_descr[counter][3]=indices[5];
    hash_mult[counter]=dummy;
  }

  // old-2010-04: if (indices[0]<0) { // End-of-Data-Set
  if (indices[1]==STOP_TERM) { // End-of-Data-Set, 2010-04 [1] == 0
    // old [0] == STOP_CFG=999999 was bad!
    term_store(id,indices,dummy);

    if (Control.output > 1 && myid == 0 && flag_out[id]) {
      printf("OUT: Cfg=%3d Term%02d\n", indices[0], indices[1]);
    }
    if (id_old != id) { id_old=id; na_out=na_in=0; } // new file
    if (Control.output > -1) {
      /* ToDo: clock output here? */
      int t2, tmin, tmax, tsum;
      t2=seconds(); // differ for each thread (collect min,max)
      tmin=tmax=tsum=t2;
      maxterms_in =nterms_in1;  /* set default for single thread computation */
      maxterms_out=nterms_out1;
#define ShiftULong (1.*(((unsigned long)1)<<(sizeof(long)*4))\
                   *1.*(((unsigned long)1)<<(sizeof(long)*4)))
      na_out=nterms_out2*ShiftULong + nterms_out1;
      na_in =nterms_in2 *ShiftULong + nterms_in1;
#if (PARALLEL_FLAG> 0 && !(EXPERIMENTAL & 2)) // do not mix pthread and MPI
 if (id != 0) { // ~GSE
      int            tmp[81+1], i;   // array used for 14*5, 81+1, use max!
      unsigned long tmp1[81+1], tmp2[81+1];
      double          d1[81+1],   d2[81+1];
      /* SUM may overflow */
      for(i=0;i<81+1;i++) tmp1[i]=numiterms[i];   // input terms
      MPI_Allreduce(tmp1,tmp2,81+1,MPI_UNSIGNED_LONG,MPI_SUM,MPI_COMM_WORLD);
      for(i=0;i<81+1;  i++) numiterms[i]=tmp2[i];   // input terms

      for(i=0;i<14*5;i++) d1[i]=tmp1[i]=numoterms[i];  // output terms
      MPI_Allreduce(tmp1,tmp2,14*5,MPI_UNSIGNED_LONG,MPI_SUM,MPI_COMM_WORLD);
      MPI_Allreduce(  d1,  d2,14*5,MPI_DOUBLE,       MPI_SUM,MPI_COMM_WORLD);
      for(i=0;i<14*5;i++) numoterms[i]=tmp2[i];   // output terms
      if (!myid)
        for(i=0;i<14*5;i++)
          if (d2[i]!=1.*numoterms[i])
            printf("Integer Overflow sum numoterms[%2d][%2d]: %13.6e\n",
                    i/5+1, i%5, d2[i]);
      if (myid>0) for(i=0;i<81+1;i++) numiterms[i]=0;
      if (myid>0) for(i=0;i<14*5;i++) numoterms[i]=0;
      MPI_Allreduce(    &t2, &tmin,  1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
      MPI_Allreduce(    &t2, &tmax,  1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
      MPI_Allreduce(    &t2, &tsum,  1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
      MPI_Allreduce(  &hcol,   tmp,  1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
      hcol=tmp[0];
      MPI_Allreduce(  &nflush, tmp,  1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
      nflush=tmp[0];
      // measure balance
      MPI_Allreduce( &nterms_in1,tmp2, 1, MPI_LONG, MPI_MAX, MPI_COMM_WORLD);
      maxterms_in=tmp2[0];
      MPI_Allreduce(&nterms_out1,tmp2, 1, MPI_LONG, MPI_MAX, MPI_COMM_WORLD);
      maxterms_out=tmp2[0];
      // sum
      d1[0]=nterms_out2*ShiftULong+nterms_out1;
      d1[1]=na_in;
      d1[2]=na_out;
      tmp1[0]=nterms_out0;  // summed to zero
      tmp1[1]=nterms_in1;
      tmp1[2]=nterms_out1;
      tmp1[3]=nterms_in2;
      tmp1[4]=nterms_out2;
      MPI_Allreduce(tmp1,tmp2,5,MPI_UNSIGNED_LONG,MPI_SUM,MPI_COMM_WORLD);
      MPI_Allreduce(  d1,  d2,3,MPI_DOUBLE,       MPI_SUM,MPI_COMM_WORLD);
      nterms_out0=tmp2[0]; // summed to zero
      nterms_in1 =tmp2[1];
      nterms_out1=tmp2[2];
      nterms_in2 =tmp2[3];
      nterms_out2=tmp2[4];
      MPI_Allreduce(&max_buff,tmp2,1,MPI_UNSIGNED_LONG,MPI_MAX,MPI_COMM_WORLD);
      // JS-Feb10: how to get max_bcfg from the thread of max_buff?
      //    if (tmp2==max_buff) send bcfg, else send 0 via Allreduce.max !?
      if (max_buff!=tmp2[0]) max_bcfg=0;
      max_buff=tmp2[0];
      MPI_Allreduce(&max_bcfg,tmp2,1,MPI_UNSIGNED_LONG,MPI_MAX,MPI_COMM_WORLD);
      max_bcfg=tmp2[0];
      if (!myid)
        if (d2[0]!=1.*nterms_out1)
          printf("Integer Overflow sum nterms_out: %13.6e\n", d2[0]);
      na_in =d2[1];
      na_out=d2[2];
 } // ~GSE
#endif
      /* output debugging infos */
      // output this as a statistic information
      if (myid==0 && flag_out[id]) {
        double sumi=0, sumo=0;
        for (i=0;i<=((id==3)?81:62);i++){
          if (numiterms[i]) printf(" i_terms%02d.%02d: %12lu  %s\n",
                        map[i], (int)i, numiterms[i],
                        ((id==3)?term81s[i]:term62s[i]));
          sumi += numiterms[i];
        }
        for (i=0;i<14*5;i++){
          /* output max 3chars of term62s (as collected term13) */
          if (numoterms[i])
            printf(" o_terms%02d:    %12lu  %3.3s [%d]\n",
              (int)i/5, numoterms[i], term13s[i/5], (int)(i%5));
          sumo += numoterms[i];
        }
        printf(" collect_term maxbuff= %8ld     max_bcfg= %6ld\n"
               " collect_term lenbuff= %8ld     hash_bits=   %3d\n",
                max_buff, max_bcfg, hash_max, hash_bits);
        printf(" collect_term hash collisions= %6d\n"
               "              buffer_flushs=   %6d\n", hcol, nflush);
        if (nterms_in1 & nterms_out1) // 1/0=nan may cause SIGFPE (Tru64-alpha) JS Oct09
        printf(" collect_term maximum unbalance in= %6.2f%% out= %6.2f%%\n",
                maxterms_in *100.*npe/nterms_in1 ,
                maxterms_out*100.*npe/nterms_out1);  // *npe fixed JS100222
        printf(" collect_terms    double        in= %12.0f  out= %10.0f\n",    na_in-((id)?npe:0), na_out);
        printf(" collect_terms    long          in= %12lu  out= %10lu\n", nterms_in1-((id)?npe:0), nterms_out1);
        printf("                  long overflow in= %12lu  out= %10lu\n", nterms_in2,     nterms_out2);
        printf("                  summed to zero   %12s   out= %10lu\n", " ", nterms_out0);
        printf(" collect_term bond summed terms in= %12.0f  out= %10.0f\n", sumi,  sumo);
        printf(" collect_term storage memory per term: 6*int+dbl=%d (or compressed -25%%)\n",
                (int)(6*sizeof(int)+sizeof(double)));
        printf(" collect_term.t[min]= %6.2f .. %6.2f .. %6.2f  npe= %d id= %d\n\n",
                tmin/60., tsum/npe/60., tmax/60., npe, id);
      }
    }
    max_buff=0;
    nterms_out2=nterms_in2=0;
    nterms_out1=nterms_in1=0;
    nterms_out0=0; // v090327, v0902 we showed wrong numbers
    hnew=hhit=hcol=0;   // reset number of collisions
    for (i=0;i<81+1;i++) numiterms[i]=0; // initialize counter
    for (i=0;i<14*5;i++) numoterms[i]=0; // initialize counter
  }
}
#else /* dont collect */
void collect_term(int indices[6],double dummy, int id) 
{
  int nio;
  const int *map;

  map=((id==3)?map81:map62); // choose the right mapping

  term_store(id,indices,dummy);
	  
  if (flag_out[id])
  if (Control.output > 1 && myid == 0 && nterms_out1 < MAX_out) {
     printf("OUT: Cfg=%3d Term%02d.%02d CCM= %4d %4d %4d %4d  % 10.7f\n",
     indices[0], map[indices[1]], indices[1],
     indices[2], indices[3],
     indices[4], indices[5], dummy);
     if (nterms_out1+1==MAX_out)
        printf("excessive output is stopped after %d terms\n", MAX_out);
  }
  // old-2010-04: if (indices[0]<0) {
  if (indices[1]==STOP_TERM) {
    /* all terms collected, obtain finished for one bond */
#if (PARALLEL_FLAG> 0 && !(EXPERIMENTAL & 2)) // do not mix pthread and MPI
      long tmp;
      MPI_Allreduce(&nterms_out1,  &tmp,  1, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
      nterms_out1=tmp;
#endif
    if (Control.output > 0 && myid == 0 && flag_out[id]) {
      /* output debugging infos */
      printf(" collect_term  terms out = %8ld\n", nterms_out1);
      printf("                overflow = %8ld\n", nterms_out2);
      nterms_out2=nterms_out1=0;
    }
  }
  nterms_out1++; if(!nterms_out1) nterms_out2++;
  nterms_in1++;  if(!nterms_in1)  nterms_in2++; /* in++ for compability */
}
#endif

/* Jun08 Joerg: collect_term2    replaced by collect_term(...,id=2) */

/* Jun08 Joerg: collect_exc_term replaced by collect_term(...,id=3) */
